local AREAMAP = {
  11, 12, 13, 14, 15, 21, 22, 23, 31, 32, 33, 34, 35, 36, 37, 41,
  42, 43, 44, 45, 46, 50, 51, 52, 53, 54, 61, 62, 63, 64, 65 
};

local BITMAP = {
  7, 9, 10, 5, 8, 4, 2, 1, 6, 3, 7, 9, 10, 5, 8, 4, 2 
};

local VAILDATE_SEQUENCE = "10X98765432";

local PATTERN15 = "^(%d%d)%d%d%d%d(%d%d)(%d%d)(%d%d)%d%d%d$"
local PATTERN18 = "^(%d%d)%d%d%d%d(%d%d%d%d)(%d%d)(%d%d)%d%d%d([X%d])$"

--函数说明
--判断指定字符串是否满足身份证规则
--参数说明
--str：身份证字符串
--返回值
--boolean true：满足规则，false：不满足规则
function verify(str)
  --print(str)
  --为空直接返回false
  if str == nil then
    return false
  end
  --如果长度不是15或18位返回false
  if #str ~= 15 and #str ~= 18 then
    return false
  end
  str = string.upper(str)
  --如果长度为15位，且不是纯数字返回false
  if #str == 15 and tonumber(str) == nil then
    return false
  end
  --如果长度为15位，且前17位不是纯数字返回false
  if #str == 18 and (tonumber(string.sub(str, 1, 17)) == nil or (tonumber(string.sub(str, 18, 18)) == nil and string.sub(str, 18, 18) ~= "X")) then
    return false
  end
  if #str == 15 then
    local area, year, month, day = str:match(PATTERN15)
    if area == nil or year == nil or month == nil or day == nil then
      return false
    end
    year = "19"..year;
    if not isVaildDateFormat(year, month, day) then
      return false
    end
    return true
  else
    local area, year, month, day, vaildate = str:match(PATTERN18)
    if area == nil or year == nil or month == nil or day == nil or vaildate == nil then
      return false
    end
    if not isVaildDateFormat(year, month, day) then
      return false
    end
    if compareDate(year, month, day, '1900', '01', '01') < 0 or compareToday(year, month, day) > 0 then
      return false
    end
    local sum = 0
    for i = 1, 17 do
      sum = sum + (BITMAP[i] * string.sub(str, i, i))
    end
    local index = (sum % 11) + 1;
    return string.sub(VAILDATE_SEQUENCE, index, index) == vaildate
  end
end

function isInAreaList(str)
  --print(str)
  if str == nil or #str ~= 2 then
    return false
  end
  for i = 1, #AREAMAP do
    --print(AREAMAP[i])
    if(tostring(AREAMAP[i]) == str) then
      return true
    end
  end
  return false
end

function compareToday(sourceYear, sourceMonth, sourceDay)
  local year = os.date("%Y", os.time())
  local month = os.date("%m", os.time())
  local day = os.date("%d", os.time())
  return compareDate(sourceYear, sourceMonth, sourceDay, year, month, day)
end

function compareDate(sourceYear, sourceMonth, sourceDay, targetYear, targetMonth, targetDay)
  if sourceYear == targetYear then
    if sourceMonth == targetMonth then
      return sourceDay - targetDay; 
    else
      return sourceMonth - targetMonth; 
    end
  else
    return sourceYear - targetYear; 
  end
end

function isVaildDateFormat(year, month, day)
  local orgDateString = year..month..day
  --print("year="..year.." month="..month.." day="..day)
  if year == nil or #year ~= 4 or month == nil or #month ~= 2 or day == nil or #day ~= 2 then
    return false
  end
  if tonumber(year) == nil or tonumber(month) == nil or tonumber(day) == nil then
    return false
  end
  local time = os.time({year=year,month=month,day=day,hour=0,min=0,sec=0})
  return os.date("%Y%m%d", time) == orgDateString
end

local function insertStr(loc, addStr, originStr)
  if loc < 0 or loc > #originStr then
    return false
  end
  local beginStr = string.sub(originStr, 1, loc)
  local endStr = string.sub(originStr, loc + 1)
  beginStr = beginStr .. addStr
  return beginStr .. endStr
end

function transID15to18(idcardNum)
  if idcardNum == nil then
    return false
  end
  if #idcardNum ~= 15 then
    return false
  end
  --step1 将15位身份证号加入出生年月到17位
  local tempSeventeenCharStr = insertStr(6, "19", idcardNum)

  --step2 将step1得到的身份证17位数分别乘以不同的系数
  --step3 将这17位数字和系数相乘的结果相加
  local sum = 0
  for i = 1, #BITMAP do
    sum = sum + (BITMAP[i] * string.sub(tempSeventeenCharStr, i, i))
  end
  --step4 将step3的结果除以11，得出余数
  local index = sum % 11 + 1
  local endIdNum = string.sub(VAILDATE_SEQUENCE, index, index)
  return insertStr(17, endIdNum, tempSeventeenCharStr)
end

-- 测试15位转18位身份证号码，最后一项是异常测试用例
local testIDInfo = {
  {src="440253850213582", dst="440253198502135821"},
  {src="231002890330208", dst="231002198903302089"},
  {src="340102821206203", dst="340102198212062039"},
  {src="340102821206203", dst="340102198212062031"}
}

-- 测试日期格式是否正确，最后一项是异常测试用例
local testDateFormat = {
  {year="2015", month="01", day="01"},
  {year="2015", month="12", day="31"},
  {year="2015", month="06", day="30"},
  {year="2015", month="02", day="29"}
}

-- 测试身份证格式是否正确，最后一项是异常测试用例
local testIDNumber = {
  "440253850213582",
  "231002890330208",
  "340102198212062039",
  "888888888888888888"
}

-- 测试用例执行总入口
local function testMain()
  --验证15位转18位身份证转换
  print("--------------transID15to18 check--------------")
  for i = 1, #testIDInfo do
    local src = testIDInfo[i]["src"]
    local result = transID15to18(src)
    if result == testIDInfo[i]["dst"] then
      print("pass")
    else
      print(string.format("not pass, target: %s, result is: %s", testIDInfo[i]["dst"], result))
    end
  end

  print("--------------isVaildDateFormat check--------------")
  for i = 1, #testDateFormat do
    local year = testDateFormat[i]["year"]
    local month = testDateFormat[i]["month"]
    local day = testDateFormat[i]["day"]
    local result = isVaildDateFormat(year, month, day)
    if result then
      print("pass")
    else
      print(string.format("not pass, year: %s month: %s day: %s", year, month, day))
    end
  end

  print("--------------verify check--------------")
  for i = 1, #testIDNumber do
    local src = testIDNumber[i]
    local result = verify(src)
    if result then
      print("pass")
    else
      print("not pass, idcard: " .. src)
    end
  end
end

-- testMain()
